from ngsolve import *
from netgen.geom2d import SplineGeometry
import math
import numpy
import vtk


def solveStatic(grid_mesh, E, nu):
    def _stress(u, mu, lam, grid_mesh):
        """
        Линейный тензор напряжения.
        """

        return 2 * mu * _epsilon(u) + lam * Trace(_epsilon(u)) * \
               Id(grid_mesh.dim)

    def _epsilon(u):
        """
        Линейный тензор деформации.
        """

        return 0.5 * (u.Deriv().trans + u.Deriv())

    coef_mu = [1 / 2 * (E[mat] / (1 + nu[mat])) for mat in grid_mesh.GetMaterials()]

    coef_lam = [E[mat] * nu[mat] / ((1 + nu[mat]) * (1 - 2 * nu[mat])) for mat in grid_mesh.GetMaterials()]

    solve_func = H1(grid_mesh, order=3, dirichlet='bc', dim=new_mesh_bc.dim)

    u, v = solve_func.TrialFunction(), solve_func.TestFunction()

    mu = CoefficientFunction(coef_mu)
    lam = CoefficientFunction(coef_lam)

    bform = BilinearForm(solve_func)
    bform += SymbolicBFI(2 * mu * InnerProduct(_epsilon(u), _epsilon(v)) + lam * Trace(u.Deriv()) * Trace(v.Deriv()))

    force_1 = CoefficientFunction((0, 1E7))

    lform = LinearForm(solve_func)
    lform += SymbolicLFI(force_1 * v, definedon=grid_mesh.Boundaries('bcf'))

    bform.Assemble()
    lform.Assemble()

    grid = GridFunction(solve_func, name='grid_1')
    grid.vec.data = bform.mat.Inverse(solve_func.FreeDofs()) * lform.vec

    displacement = grid
    deformation = _epsilon(grid)
    stress = _stress(grid, mu, lam, grid_mesh)

    return deformation, displacement, stress


def getDisplacement(displacement):

    array_displacement = numpy.array([])

    for dof in grid_mesh.vertices:
        displacement_dof = math.sqrt(displacement(grid_mesh(*dof.point))[0] * displacement(grid_mesh(*dof.point))[0] +
                                     displacement(grid_mesh(*dof.point))[1] * displacement(grid_mesh(*dof.point))[1])

        array_displacement = numpy.append(array_displacement, displacement_dof)

    return array_displacement


def getStress(stress):
    array_stress = numpy.array([])

    for dof in grid_mesh.vertices:
        stress_dof = math.sqrt(stress(grid_mesh(*dof.point))[0] * stress(grid_mesh(*dof.point))[0] -
                               stress(grid_mesh(*dof.point))[0] * stress(grid_mesh(*dof.point))[3] +
                               stress(grid_mesh(*dof.point))[3] * stress(grid_mesh(*dof.point))[3] +
                               3 * stress(grid_mesh(*dof.point))[1] * stress(grid_mesh(*dof.point))[1])

        array_stress = numpy.append(array_stress, stress_dof)

    return array_stress


def getDeformation(deformation):
    array_deformation = numpy.array([])

    for dof in grid_mesh.vertices:
        deformation_dof = (2 / (3 * math.sqrt(2))) * \
                          math.sqrt((deformation(grid_mesh(*dof.point))[0] - deformation(grid_mesh(*dof.point))[3]) *
                                    (deformation(grid_mesh(*dof.point))[0] - deformation(grid_mesh(*dof.point))[3]) +
                                    deformation(grid_mesh(*dof.point))[0] * deformation(grid_mesh(*dof.point))[0] +
                                    deformation(grid_mesh(*dof.point))[3] * deformation(grid_mesh(*dof.point))[3] +
                                    6 * deformation(grid_mesh(*dof.point))[2] * deformation(grid_mesh(*dof.point))[2])

        array_deformation = numpy.append(array_deformation, deformation_dof)

    return array_deformation


geo = SplineGeometry()
geo.Load('bone.in2d')

grid_mesh = Mesh(geo.GenerateMesh(maxh=10))

new_mesh_bc = grid_mesh.ngmesh
new_mesh_bc.SetBCName(1, 'bc')
new_mesh_bc.SetBCName(3, 'bcf')

new_mesh_bc.SetMaterial(1, 'mat1')

E = {'mat1': 1e10, 'mat2': 5e6, 'mat3': 1.14e11}
nu = {'mat1': 0.3, 'mat2': 0.45, 'mat3': 0.3}

deformation, displacement, stress = solveStatic(grid_mesh, E, nu)

print(getDisplacement(displacement))